;;; lisp file for xyzzy ver0.2.2.233
;;; essay-mode.l ver0.1.3.0.8 // crayonzen [2006-06-17]23:10:38(+0900)

(provide "essay-mode")

(in-package "editor")

(export '(essay-mode
	  essay-newline
	  essay-newline-and-indent
	  indent-follow
	  indent-follow-forward

	  essay-indent-region
	  essay-top-indent
	  essay-delete-comment
	  *essay-mode-map*
	  *essay-mode-syntax-table*
	  *essay-mode-hook*

	  delete-backward-spaces
	  delete-forward-spaces
	  delete-line-top-spaces
	  delete-line-end-spaces
	  ))

(defvar *essay-mode-hook* nil)
(defvar *essay-mode-map* nil)
(defvar-local *essay-mode* nil)
(defvar-local *keep-top-indent* 2)

(unless *essay-mode-map*
  (setq *essay-mode-map* (make-sparse-keymap))
  (define-key *essay-mode-map* #\LFD 'essay-newline)
  (define-key *essay-mode-map* #\RET 'essay-newline-and-indent)
  (define-key *essay-mode-map* #\C-F12 'essay-indent-region)
  (define-key *essay-mode-map* #\TAB 'indent-follow)
  (set-extended-key-translate-table exkey-S-tab #\F20)
  (define-key *essay-mode-map* #\F20 'indent-follow-forward)
  (set-extended-key-translate-table exkey-C-tab #\F21)
  (define-key *essay-mode-map* #\F21 'let-tab)
  )

(defvar *essay-mode-syntax-table* nil)
(unless *essay-mode-syntax-table*
  (setq *essay-mode-syntax-table* (make-syntax-table))
;  (set-syntax-symbol *essay-mode-syntax-table* #\&)
;  (set-syntax-symbol *essay-mode-syntax-table* #\*)
;  (set-syntax-symbol *essay-mode-syntax-table* #\+)
;  (set-syntax-symbol *essay-mode-syntax-table* #\-)
;  (set-syntax-symbol *essay-mode-syntax-table* #\/)
;  (set-syntax-symbol *essay-mode-syntax-table* #\<)
;  (set-syntax-symbol *essay-mode-syntax-table* #\=)
;  (set-syntax-symbol *essay-mode-syntax-table* #\>)
;  (set-syntax-symbol *essay-mode-syntax-table* #\|)
  (set-syntax-symbol *essay-mode-syntax-table* #\_))


(defvar *essay-mode-abbrev-table* nil)
(unless *essay-mode-abbrev-table*
  (define-abbrev-table '*essay-mode-abbrev-table*))

(defun essay-newline (&optional (arg 1))
  (interactive "*p")
  (newline arg)
  )

(defun essay-newline-and-indent (&optional (arg 1))
  (interactive "*p")
  (if overwrite-mode
      (newline arg)
    (progn
      (delete-backward-spaces)
      (newline arg)
      (essay-indent)))
  )

;;  (insert #\LFD arg)
;;  (let ((*last-command-char* #\LFD))
;;    (self-insert-command arg))
;;  (let ((*last-command-char* #\RET))
;;    (self-insert-command arg))

;;real indent
(defun essay-indent ()
  (interactive "*")
  (if (not (check-line-empty -1))
      (progn
	(save-excursion
	  (forward-line -1)
	  (skip-chars-forward " \t@")
	  (setq lfdcolumn (current-column)))

	;(save-excursion
	;(delete-line-top-spaces))
	;;comment out;; keep for come along with a right space

	(goto-bol)
	(setq i (cond
		 ((<= lfdcolumn *keep-top-indent*)
		  0)
		 (t
		  lfdcolumn)))
	(indent-to i))
    )
  )

;;again indent
(defun essay-indent-line ()
  (interactive "*")
  (if (not (check-line-empty -1))
      (progn
	(save-excursion
	  ;current column
	  (goto-bol)
	  (skip-chars-forward " \t@")
	  (setq curcolumn (current-column))
	  ;LFD column
	  (forward-line -1)
	  (skip-chars-forward " \t@")
	  (setq lfdcolumn (current-column)))

	(save-excursion
	  (setq fspaces (- (current-column) ;; = 0
			   (progn
			     (skip-chars-forward " \t@")
			     (current-column))))
	  (delete-line-top-spaces))

	(goto-bol)
	(setq i (cond
		 ((= curcolumn 0)
		  0)
		 ((and (> curcolumn 0)
		       (<= curcolumn *keep-top-indent*))
		  *keep-top-indent*)

		 ((and (> curcolumn *keep-top-indent*)
		       (<= lfdcolumn *keep-top-indent*))
		  curcolumn)
		 (t
		  lfdcolumn)))
	(indent-to i))
    )
  )

;;curcolumn > 0
#|
((and (> curcolumn 0) (<= curcolumn *keep-top-indent*))
 *keep-top-indent*)
((and (> curcolumn *keep-top-indent*) (<= lfdcolumn *keep-top-indent*))
 curcolumn)
((and (> curcolumn *keep-top-indent*) (> lfdcolumn *keep-top-indent*))
 lfdcolumn)
|#

(defun check-line-empty (&optional (arg))
  (interactive)
  (save-excursion
    (if (= 0 arg)
	(progn
	  (goto-bol)
	  (skip-chars-forward " \t@")
	  (eolp))
      (progn
	(setq narg (forward-line arg))
	(if (eq narg arg)
	    (progn
	      (skip-chars-forward " \t@")
	      (eolp))
	  t)))
    )
  )

(defun essay-top-indent (n1)
  (interactive "ntop indent:")
  (essay-modify-indent n1)
  (setq *keep-top-indent* n1)
  (message "set keep top indent: ~S" n1)
  )

(defun essay-modify-indent (n1)
  (interactive "*")
  (long-operation
    (when (/= n1 *keep-top-indent*)
      (setq n2 (- n1 *keep-top-indent*))
      (save-excursion
	(goto-char (point-min))
	(delete-region (progn (goto-bol) (point))
		       (progn
			 (skip-chars-forward " \t@")
			 (setq curcolumn (current-column))
			 (point)))
	(when (/= curcolumn 0)
	  (indent-to (+ curcolumn n2)))

	(while (forward-line 1)
	  (delete-region (progn (goto-bol) (point))
			 (progn
			   (skip-chars-forward " \t@")
			   (setq curcolumn (current-column))
			   (point)))
	  (when (/= curcolumn 0)
	    (indent-to (+ curcolumn n2)))))))
  )

;;;;;;;;;;;;;;;; keyword file (in "etc" folder)
(export '(*essay-keyword-file*
	  essay-load-keyword-file
	  *essay-keyword-formats*
	  ))

(defvar *essay-keyword* nil)
(defvar *essay-regexp-keyword* nil)

(defvar *essay-keyword-hash-table* nil)
(defvar *essay-keyword-file* "essay")

(defun essay-load-keyword-file ()
  (let* ((path (directory-namestring (merge-pathnames "xyzzy.exe" (si:system-root))))
	 (path (merge-pathnames "etc" path))
	 (readfile (merge-pathnames *essay-keyword-file* path)))
    (when (file-exist-p readfile)
      (make-local-variable 'highlight-keyword)
      (setq highlight-keyword t)
      (and (null *essay-keyword-hash-table*)
	   (setq *essay-keyword-hash-table*
		 (load-keyword-file *essay-keyword-file* t)))
      (when *essay-keyword-hash-table*
	(make-local-variable 'keyword-hash-table)
	(setq keyword-hash-table *essay-keyword-hash-table*)))
    (make-local-variable 'regexp-keyword-list)
    (setq regexp-keyword-list *essay-keyword-formats*))
  )
;;;;;;;;;;;;;;;; keyword regexp
;;ipj 0123456789 :F ܰ3
;;@V{ | :F ܰ3
;;@Rg s܂͍s // :F  ܰ1

(defvar *essay-keyword-formats*
  (compile-regexp-keyword-list
   '(
     ;("\\([0-9]\\)" t (:keyword 9) t) ;
     ("\\([0-9]\\)" t (:keyword :string :bold) :string)
     ("\\([0-9]\\)" t (:keyword :tag :bold) :tag)
     ("\\([0-9]\\)" t (:keyword :comment :bold) :comment)
     ("\\([0-9]\\)" t (:keyword 9))

     ;("\|" t (:keyword 2 :bold) t)
     ("\|" t (:keyword :string :bold) :string)
     ("\|" t (:keyword :tag :bold) :tag)
     ("\|" t (:keyword :comment :bold) :comment)
     ("\|" t (:keyword 2 :bold))

     ("^//" t (:keyword :comment :line) t) ;
     ("//\$" t (:keyword :comment :line) t) ;
     )))

;;;wR[h@wF@QlF
;;p :F ܰ3
;;ȊỎR[hSāipj:F 
;;pEpEȊỎR[hip) :F ܰ3
;("\\([--]\\)" t (:keyword 2) t) ;p
;("\\([!-/:-]\\)" t (:keyword :string) t) ;ȊỎR[hSāipj
;("\\([ -/:-@[-\`{-]\\)" t (:keyword 2) t) ;pEpEȊỎR[hip)

(defun essay-delete-comment ()
  (interactive "*")
  (when (yes-or-no-p "čs폜܂H~%is܂͍s//̂sj")
    (let (start
	  end)
      (cond ((pre-selection-p)
	     (setq start (selection-mark) end (selection-point)))
	    ((mark t)
	     (setq start (mark) end (point)))
	    (t
	     (setq start (point-min) end (point-max))))
      (if (> start end)
	  (rotatef start end))
      (save-excursion
	(save-restriction
	  (narrow-to-region (progn
			      (goto-char start) (goto-bol)
			      (point))
			    (progn
			      (goto-char end) (goto-eol)
			      (point)))
	  (goto-char (point-min))
	  (while (scan-buffer "^//" :regexp t)
	    (delete-region (point)
			   (progn
			     (goto-eol)
			     (+ 1 (point)))))
	  (goto-char (point-min))
	  (while (scan-buffer "//\$":regexp t)
	    (delete-region (progn
			     (goto-bol)
			     (point))
			   (progn
			     (goto-eol)
			     (+ 1 (point)))))))))
  )

;;;;;;;;;;;;;;;; tool
(defun delete-backward-spaces ()
  (interactive "*")
  (delete-region (point)
		 (progn
		   (skip-chars-backward " \t@")
		   (point))))

(defun delete-forward-spaces ()
  (interactive "*")
  (delete-region (point)
		 (progn
		   (skip-chars-forward " \t@")
		   (point))))

(defun delete-line-top-spaces ()
  (interactive "*")
  (delete-region (progn (goto-bol) (point))
		 (progn (skip-chars-forward " \t@") (point)))
  )

(defun delete-line-end-spaces ()
  (interactive "*")
  (delete-region (progn (goto-eol) (point))
		 (progn (skip-chars-backward " \t@") (point)))
  )

(defun essay-indent-region ()
  (interactive "*")
  (cond ((pre-selection-p)
	 (essay-indent-region-line (selection-mark) (selection-point)))
	((mark t)
	 (essay-indent-region-line (mark nil) (point)))
	(t
	 (essay-indent-region-line (point-min) (point-max))))
  )

(defun essay-indent-region-line (from to)
  (interactive "*r")
  (if (> from to)
      (rotatef from to))
  (save-excursion
    (save-restriction
      (narrow-to-region (progn (goto-char from) (goto-bol) (point))
			(progn (goto-char to) (goto-eol) (point)))
      (goto-char (point-min))
      (delete-line-end-spaces)
      (while (forward-line 1)
	(delete-line-end-spaces)
	(unless (bolp)
	  (funcall mode-indent-line-command)))))
  t)

;;;;;;;;;;;;;;;; tab
(defun let-tab (&optional (arg 1))
  (interactive "*p")
  (let ((*last-command-char* #\TAB))
    (self-insert-command arg))
  )

(defun indent-follow (&optional fflag (arg 1))
  (interactive "*p")
  (if overwrite-mode
      (indent-follow-overwrite-mode fflag arg)
    (indent-follow-common fflag arg)))

(defun indent-follow-overwrite-mode (&optional fflag (arg 1))
  (if fflag
      (if (bolp)
	  (forward-char -1)
	(progn
	  (skip-chars-backward " \t@")
	  (while (not (bolp))
	    (if (or (char= (preceding-char) #\TAB)
		    (char= (preceding-char) (character " "))
		    (char= (preceding-char) (character "@")))
		(return)
	      (forward-char -1)))))
    (if (eolp)
	(let-tab arg)
      (progn
	(while (not (eolp))
	  (if (or (char= (following-char) #\TAB)
		  (char= (following-char) (character " "))
		  (char= (following-char) (character "@")))
	      (return)
	    (forward-char)))
	(skip-chars-forward " \t@")
	)))
  )

(defun indent-follow-common (&optional fflag (arg 1))
  (save-excursion
    (setq cc (current-column))
    (if fflag
	(progn
	  (goto-eol)
	  (skip-chars-forward " \t\n@"))
      (progn
	(goto-bol)
	(skip-chars-backward " \t\n@")))
    (goto-column cc)
    (while (not (eolp))
      (if (or (char= (following-char) #\TAB)
	      (char= (following-char) (character " "))
	      (char= (following-char) (character "@")))
	  (return)
	(forward-char)))
    (skip-chars-forward " \t@")
    (if (and (not (eolp)) (< cc (current-column)))
	(setq indentcolumns (current-column))
      (setq indentcolumns 0)))

  (progn
    (if (= 0 indentcolumns)
	(let-tab arg)
      (progn
	(if (pre-selection-p)
	    (delete-region (selection-mark) (selection-point)))
	(indent-to indentcolumns))))
  )

(defun indent-follow-forward (&optional (arg 1))
  (interactive "*")
  (indent-follow t arg)
  )

;;;;;;;;;;;;;;;; mode
(defun essay-mode ()
  (interactive)
  (kill-all-local-variables)
  (setq buffer-mode 'essay-mode)
  (setq mode-name "Essay")
  (use-keymap *essay-mode-map*)
  (use-syntax-table *essay-mode-syntax-table*)
  (setq *local-abbrev-table* *essay-mode-abbrev-table*)
  (make-local-variable 'mode-indent-line-command)
  (setq mode-indent-line-command 'essay-indent-line)
  (essay-load-keyword-file)
  (run-hooks '*essay-mode-hook*))
