;;; lisp file for xyzzy ver0.2.2.234
;;; free-cursor-mode.l ver.0.0.2.4 // crayonzen [2006-12-10]13:40:13(+0900)

;; how-to: load
;;       : write in .xyzzy or siteinit.l
;;
;; (load-library "free-cursor-mode")

;; & (overwrite-mode t) ?

(provide "free-cursor-mode")

(in-package "editor")

(export '(free-cursor-mode
	  *free-cursor-mode-map*
	  ))

(defvar-local *last-click-line-number* nil)

(defvar-local *free-cursor-mode* nil)
(defvar *free-cursor-mode-map* nil)
(unless *free-cursor-mode-map*
  (setq *free-cursor-mode-map* (make-sparse-keymap))
  (define-key *free-cursor-mode-map* #\M-S 'center-paragraph)

  (define-key *free-cursor-mode-map* #\LBtnDown 'free-cursor)
  ;;  (define-key *free-cursor-mode-map* #\LBtnUp'free-cursor-mouse-nop)

  ;;  (define-key *free-cursor-mode-map* #\C-RBtnDown 'free-cursor)
  ;;  (define-key *free-cursor-mode-map* #\C-RBtnUp 'free-cursor-mouse-nop)

  (define-key *free-cursor-mode-map* #\Down 'free-cursor-next-line)
  (define-key *free-cursor-mode-map* #\Up 'free-cursor-previous-line)
  (define-key *free-cursor-mode-map* #\Left 'free-cursor-backward-char)
  (define-key *free-cursor-mode-map* #\Right 'free-cursor-forward-char)
  (define-key *free-cursor-mode-map* #\C-Down 'free-cursor-fast-scroll-up)
  (define-key *free-cursor-mode-map* #\C-Up 'free-cursor-fast-scroll-down)

  (define-key *free-cursor-mode-map* #\PageDown 'free-cursor-next-page)
  (define-key *free-cursor-mode-map* #\PageUp 'free-cursor-previous-page)
  (define-key *free-cursor-mode-map* #\C-l 'free-cursor-recenter)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;mouse

(defun free-cursor ()
  (interactive)
  (mouse-left-press)
  (clear-free-cursor-line *last-click-line-number*)
  (non-tab)

  ;current-virtual-line-number
  ;current-virtual-column
  (let ((insertcount
	 (save-excursion
	   (goto-char (point-max))
	   (- *last-mouse-line* (current-line-number)))))

    (if (< 0 insertcount)
	(progn
	  (goto-char (point-max))
	  (insert #\LFD (incf insertcount))
	  (forward-line -1)))

    ;(message "test: ~S@:~S" *last-mouse-column* (current-line-number))
    (let (insertlength
	  followingchar)
      (save-excursion
	(setq insertlength (- *last-mouse-column* (progn (goto-eol) (current-column))))
	(setq followingchar (progn (goto-column *last-mouse-column*) (following-char))))
      (if (< 0 insertlength)
	  (progn
	    (goto-eol)
	    (insert #\SPC insertlength)
	    ))
      (if (and (char= followingchar (character "@"))
	       (/= *last-mouse-column* (current-column)))
	  (progn
	    (goto-column *last-mouse-column*)
	    (insert #\SPC 2)
	    (delete-char)
	    (forward-char -1))
	)
      )
    )

  (setq *last-click-line-number* (current-line-number))
  (set-goal-column (max 0 *last-mouse-column*))
  )

;;(defun free-cursor-mouse-nop ()
;;  (interactive)
;;  (continue-pre-selection)
;;  (selection-leave)
;;  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;key

;(following-char) ;EOF #\NULL
;(goalcolumn (goal-column)) ;(set-goal-column goalcolumn)

(defun free-cursor-next-line (&optional (n 1))
  (interactive "p")
  (let ((goalcolumn (goal-column)))
    (if (or (< 0 n)
	    (/= 1 (current-line-number)))
	(clear-free-cursor-line *last-click-line-number*))
    (if (null (forward-line n))
	(if (< 0 n)
	    (progn
	      (insert #\LFD)
	      (insert #\SPC goalcolumn)))
      (free-cursor-goto-column goalcolumn))
    (setq *last-click-line-number* (current-line-number))
    (set-goal-column goalcolumn)
    )
  )

(defun free-cursor-previous-line (&optional (n 1))
  (interactive "p")
  (free-cursor-next-line (- n))
  )

(defun free-cursor-forward-char (&optional (n 1))
  (interactive "p")
  (if (/= *last-click-line-number* (current-line-number))
      (clear-free-cursor-line *last-click-line-number*))
  (if (eolp)
      (insert #\SPC n)
    (forward-char n))
  (setq *last-click-line-number* (current-line-number))
  )

(defun free-cursor-backward-char (&optional (n -1))
  (interactive "p")
  (if (/= *last-click-line-number* (current-line-number))
      (clear-free-cursor-line *last-click-line-number*))
  (if (null (bolp))
      (progn
	(forward-char n)
	(setq currentpoint (point))
	(save-excursion
	  (delete-region (progn (goto-eol) (point))
			 (max currentpoint
			      (progn (skip-chars-backward " \t") (point)))
			 ))))
  (setq *last-click-line-number* (current-line-number))
  )

(defun free-cursor-fast-scroll-up ()
  (interactive)
  (let ((goalcolumn (goal-column)))
    (clear-free-cursor-line *last-click-line-number*)
    (fast-scroll-up)
    (free-cursor-goto-column goalcolumn)
    (setq *last-click-line-number* (current-line-number)))
  )

(defun free-cursor-fast-scroll-down ()
  (interactive)
  (let ((goalcolumn (goal-column)))
    (clear-free-cursor-line *last-click-line-number*)
    (fast-scroll-down)
    (free-cursor-goto-column goalcolumn)
    (setq *last-click-line-number* (current-line-number)))
  )

(defun free-cursor-next-page (&optional (arg 1))
  (interactive)
  (let ((goalcolumn (goal-column)))
    (clear-free-cursor-line *last-click-line-number*)
    (next-page arg)
    (free-cursor-goto-column goalcolumn)
    (setq *last-click-line-number* (current-line-number))
    )
  )

(defun free-cursor-previous-page ()
  (interactive)
  (free-cursor-next-page -1))

(defun free-cursor-recenter (&optional (arg nil))
  (interactive)
  (let ((goalcolumn (goal-column)))
    (clear-free-cursor-line *last-click-line-number*)
    (recenter arg)
    (free-cursor-goto-column goalcolumn)
    (setq *last-click-line-number* (current-line-number))
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;tool

(defun clear-free-cursor-line (&optional (linenumber (current-line-number)))
  (interactive)
  (if linenumber
      (save-excursion
	(goto-line linenumber)
	(delete-region (progn (goto-eol) (point))
		       (progn (skip-chars-backward " \t") (point)))))
  )

(defun free-cursor-goto-column (column)
  (if column
      (progn
	(setq n (- column (progn (goto-eol) (current-column))))
	(if (< 0 n)
	    (insert #\SPC n)
	  (progn
	    (non-tab)
	    (goto-column column)))))
  )

(defun non-tab ()
  (let ((tabcolumns (tab-columns (selected-buffer))))
    (save-excursion
      (save-restriction
	(narrow-to-region (progn (goto-bol) (point))
			  (progn (goto-eol) (point)))
	(goto-char (point-min))
	(while (scan-buffer "\t" :regexp t)
	  (setq insertcount (- tabcolumns (rem (current-column) tabcolumns)))
	  (insert #\SPC insertcount)
	  (delete-char)
	  )
	)))
  )

;;toggle
(defun free-cursor-mode (&optional (arg nil sv))
  (interactive "p")
  (toggle-mode '*free-cursor-mode* arg sv)
  (update-mode-line t)
  (if *free-cursor-mode*
      (set-minor-mode-map *free-cursor-mode-map*)
    (progn
      (clear-free-cursor-line)
      (unset-minor-mode-map *free-cursor-mode-map*)))
  t)

(pushnew '(*free-cursor-mode* . "FreeCursor") *minor-mode-alist* :key #'car)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;(global-set-key #\LBtnDown 'free-cursor)
;;(global-set-key #\LBtnUp 'free-cursor-mouse-nop)

;;(global-set-key #\C-RBtnDown 'free-cursor)
;;(global-set-key #\C-RBtnUp 'free-cursor-mouse-nop)
;;
;;(global-set-key #\Down 'free-cursor-next-line)
;;(global-set-key #\Up 'free-cursor-previous-line)
;;(global-set-key #\Left 'free-cursor-backward-char)
;;(global-set-key #\Right 'free-cursor-forward-char)
;;(global-set-key #\C-Down 'free-cursor-fast-scroll-up)
;;(global-set-key #\C-Up 'free-cursor-fast-scroll-down)
;;
;;(global-set-key #\PageDown 'free-cursor-next-page)
;;(global-set-key #\PageUp 'free-cursor-previous-page)
;;(global-set-key #\C-l 'free-cursor-recenter)
